/* $Id: parsehtml.c,v 1.12 2002/10/21 23:26:43 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/*** \file
The name of the file is rather misleading. Actually it has nothing to
do with parsing HTML. Instead, it contains a set of routines to operate
on internal representation of markup elements, called \em[chunks].
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stddef.h>
 #include <string.h>
#include "util.h"
#include "dstring.h"
/**** \dev\include[parsehtml.h]\restore */
#include "parsehtml.h"

// static int counter;

/*** \function
\returns a new empty chunk
*/
html_chunk *alloc_html_chunk(void)
{
	html_chunk *hc = ALLOC(html_chunk);
	hc->type = hct_none;
	hc->is_dynamic = 1;
	// counter++;
	return hc;
}

/*** \function
\returns a copy of \var[src]
*/
html_chunk *copy_html_chunk(html_chunk *src)
{
	html_chunk *hc; 
	html_param *pi, *np;

	if(!src)
		return NULL;

	hc = alloc_html_chunk();
	hc->type = src->type;
	if(hc->type == hct_char)
		hc->x.ch = src->x.ch;
	else if(hc->type == hct_tag)
	{
		hc->x.name = ds_fix(ds_copy(src->x.name));
		for(pi = src->params; pi; pi = list_next(pi))
		{
			np = ALLOC(html_param);
			np->name = ds_fix(ds_copy(pi->name));
			if(pi->val)
				np->val = ds_fix(ds_copy(pi->val));
			hc->params = list_add(hc->params, np);
		}
		hc->params = list_reversip(hc->params);
	}
	return hc;
}

static int find_htmlparam(html_param *param, dstring *key)
{
	return !ds_compare(param->name, key, ds_p_casefold, NULL);
}

static int find_htmlparam2(html_param *param, char *key)
{
	return !strcasecmp(DS_BODY(param->name), key);
}

/*** \function
\returns a value for an attribute of \var[chunk] named \var[name];
NULL if the chunk is not a tag or if it has no \var[name].
If the value is empty, an empty string is returned
*/
char *get_html_param(html_chunk *chunk, char *name)
{
    html_param *iter;
    if(chunk->type != hct_tag)
		return NULL;
	iter = list_find(chunk->params, name, (list_predicate_t)find_htmlparam2);
	if(!iter)
		return NULL;
   	return iter->val ? DS_BODY(iter->val) : "";
}

/*** \function
sets the value of an attribute \var[key] of the \var[chunk] to \var[val];
if the chunk has no such attribute, it is added;
\note \var[val] is fixed, not copied, as well as \var[key] when an attribute is
added
*/
void set_html_param(html_chunk *chunk, dstring *key, dstring *val)
{
	html_param *iter;
    if(chunk->type != hct_tag)
		return;
	iter = list_find(chunk->params, key, (list_predicate_t)find_htmlparam);
	if(iter)
	{
		if(!val)
		{
			if(iter->val)
				ds_unfix(iter->val);
			iter->val = NULL;
		}
		else
		{
			if(iter->val)
				ds_assign(iter->val, val);
			else iter->val = ds_fix(val);
		}
    }
	else
	{
		iter = ALLOC(html_param);
		iter->name = ds_fix(key);
		iter->val = val ? ds_fix(val) : NULL;
		chunk->params = list_append(chunk->params, iter);
	}
}

/**** \function
removes an attribute \var[key] from a chunk \var[hc]. If \var[hc] had no such
attribute, nothing happens
*/
void remove_html_param(html_chunk *hc, dstring *key)
{
	hc->params = list_remove(hc->params, key, (list_predicate_t)find_htmlparam, (list_destroy_t)destroy_html_param, sizeof(html_param));
}

/**** \function
wipes out all the attributes of \var[hc]
*/
void clear_html_params(html_chunk *hc)
{
	list_delete(hc->params, (list_destroy_t)destroy_html_param, sizeof(html_param));
	hc->params = NULL;
}

/**** \function
sets the name of the \var[chunk] to \var[newname]. Changes the type of \var[chunk] to
`tag'.
*/
void set_tag_name(html_chunk *chunk, dstring *newname)
{
	if(chunk->type == hct_tag)
		ds_unfix(chunk->x.name);
	else
		chunk->type = hct_tag;
	chunk->x.name = ds_fix(newname);
}

/**** \function
\returns the name of the \var[chunk]; NULL if \var[chunk] is not a tag
*/
char *get_tag_name(html_chunk *chunk)
{
	if(chunk->type != hct_tag)
		return NULL;
	return DS_BODY(chunk->x.name);
}

/**** \function
\returns a character value of the \var[chunk] if it is a character chunk;
\code[HTML_NO_CHAR] otherwise
*/
int get_html_char(html_chunk *chunk)
{
	return chunk->type == hct_char ? chunk->x.ch : HTML_NO_CHAR;
}

/**** \function
sets the character value of \var[hc] to \var[ch]. Changes the type of the chunk to
`character'
*/
void set_html_char(html_chunk *hc, int ch)
{
	if(hc->type == hct_tag)
	{
		ds_unfix(hc->x.name);
		clear_html_params(hc);
	}
	hc->type = hct_char;
	hc->x.ch = ch;
}

/**** \function
obvious
*/
void destroy_html_param(html_param *parm)
{
	ds_unfix(parm->name);
	if(parm->val)
		ds_unfix(parm->val);
}

/*** \function
sets the type of \var[hc] to null
*/
void reset_html_chunk(html_chunk *hc)
{
	if(hc->type == hct_tag)
	{
		ds_unfix(hc->x.name);
		clear_html_params(hc);
	}
	hc->type = hct_none;
}

/*** \function
deletes the \var[chunk];
frees all the dynamically allocated memory occupied by the \em[chunk].
*/
void release_html_chunk(html_chunk *hc)
{
	if(!hc)
		return;
//	counter--;
	reset_html_chunk(hc);
	if(hc->is_dynamic)
		FREE(hc);
}


