/* $Id: sl.c,v 1.34 2003/01/02 22:40:49 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/// \section[COMMAND LINE OPTIONS]
// \em[-i infile] -- name of the standard input stream file (none by default)
// 
// \em[-o outfile] -- name of the standard output stream file (console by
// default); \em[none] means no output stream opened; the same effect may be
// achieved by using \em[-nout] as a part of the program name.
//
// \em[-d preload-modules] -- name of an alternative file containing the
// list preloaded modules (by default, preload.scm in \em2[STD_LOCATION][)]
//
// \em[-l location] -- path to alternative location for NSL system files,
// (\em[STD_LOCATION] by default)
//
// \em[-s script-path] -- a path added to a path list for scripts
//
// \em[-c] -- run sl as a CGI client. The same effect may be achieved by
// using \em[cgi] as a part of the program name
//
// \em[-f] -- using flat layout, that is, scripts and automata are considered to be
// at the top of the standard location (useful for running without installation)
//
// \em[-g reserved-limit] -- the amount of memory kept by a garbage collector for reuse
//
// \em[-G garbage-limit] -- the amount of garbage memory which 
// triggers garbage collector.
//
// Amounts are in bytes; 'k' and 'm' suffixes may be used.
// \note The latter two options had different meaning in pre-0.9 versions,
//  due to a different memory management.
// \noindent
// 
// \em[-x count]  -- the number of garbage items which triggers garbage collector
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "util.h"
#include "dstring.h"
#include "interp.h"
#include "bstream.h"
#include "cgistream.h"
#ifdef USE_LTDL
#include "modules.h"
#endif
#include "trap.h"

extern int program_argc;
extern char **program_argv;
char *nsl_location = NULL;

static stream *outs, *ins;
static int no_out;
static int as_cgi;
static int syntax_only;
static int flat_layout;
static char *def_mod;
static char *location = STD_LOCATION;


static char *add_script_path;
// static int from_cur_dir;

static void usage(char *argv0)
{
	fprintf(stderr, "Usage: %s [-fc] [-l location] [-s script-path] [-d preload-file] [-g reserved-limit] [-G garbage-limit] [-i infile] [-o outfile] script-file [arguments ...]\n", argv0);
}

static void print_version(void)
{
	puts("Tensile " VERSION);
	puts("Copyright (C) 2001, 2002 Artem V. Andreev");
	puts("Tensile comes with ABSOLUTELY NO WARRANTY.");
	puts("Tensile is a free software; it may be redistributed under the terms of\n"
		 "the GNU General Public License. See COPYING on these matters.");
}

static int get_amount(char *str)
{
	int am = strtoul(str, &str, 10);
	if(*str == 'k')
	{
		am *= 1024;
		str++;
	}
	else if(*str == 'm')
	{
		am *= 1024 * 1024;
		str++;
	}
	if(*str || am <= 0)
		fatal_error("invalid amount");
	return am;
}

typedef struct option_info
{
	char *longname;
	int has_arg;
	void *unused;
	int shortname;
	char *docstring;
} option_info;

typedef struct option_iter
{
	char **argv;
	char *inoption;
	char *arg;
	option_info *opt;
} option_iter;

static option_info *find_long_option(option_info *options, char *name)
{
	for(; options->longname; options++)
	{
		if(!strcmp(options->longname, name))
			return options;
	}
	return NULL;
}

static option_info *find_short_option(option_info *options, int ch)
{
	for(; options->longname; options++)
	{
		if(options->shortname == ch)
			return options;
	}
	return NULL;
}

static int long_getopt(option_info *options, option_iter *current)
{
	static char buffer[256];

	if(!*current->argv || **current->argv != '-' || !(*current->argv)[1])
		return -1;

	if(!current->inoption)
	{
		if((*current->argv)[1] != '-')
			current->inoption = *current->argv + 1;
		else
		{
			if(!(*current->argv)[2])
			{
				current->argv++;
				return -1;
			}
			strncpy(buffer, (*current->argv) + 2, sizeof(buffer) - 1);
			buffer[sizeof(buffer) - 1] = '\0';
			current->arg = strchr(buffer, '=');
			if(current->arg)
			{
				*current->arg = '\0';
				current->arg++;
			}
			current->opt = find_long_option(options, buffer);
			if(!current->opt || !!current->opt->has_arg != !!current->arg)
				return '?';
			current->argv++;
			return current->opt->shortname;
		}
	}
	current->opt = find_short_option(options, *current->inoption);
	if(!current->opt)
		return '?';
	if(!current->opt->has_arg)
	{
		current->inoption++;
		if(!*current->inoption)
		{
			current->inoption = NULL;
			current->argv++;
		}
	}
	else
	{
		if(current->inoption[1])
		{
			current->arg = current->inoption + 1;
			current->argv++;
		}
		else if(!current->argv[1])
			return '?';
		else
		{
			current->arg = current->argv[1];
			current->argv += 2;
		}
		current->inoption = NULL;
	}
	return current->opt->shortname;
}

static char *unknown_option(option_iter *current)
{
	static char buffer[3] = "-?";
	if(current->inoption)
	{
		buffer[1] = *current->inoption;
		return buffer;
	}
	else return *current->argv;
}

static void parse_options(int argc, char *argv[])
{
	static option_info sl_long_options[] = {{"cgi", 0, NULL, 'c'},
							  {"help", 0, NULL, 'h'},
							  {"usage", 0, NULL, 'u'},
							  {"flat", 0, NULL, 'f'},
							  {"input", 1, NULL, 'i'},
							  {"output", 1, NULL, 'o'},
							  {"location", 1, NULL, 'l'},
							  {"scriptpath", 1, NULL, 's'},
							  {"preload", 1, NULL, 'd'},
							  {"reserved-limit", 1, NULL, 'g'},
							  {"garbage-limit", 1, NULL, 'G'},
							  {"garbage-count", 1, NULL, 'x'},
							  {"version", 0, NULL, 'V'},	
							  {"syntax", 0, NULL, 'S'},
							  {NULL, 0, NULL, 0}
	};
	int c;
	option_iter current = {NULL};

	if(strstr(argv[0], "cgi"))
		as_cgi = 1;
	if(strstr(argv[0], "-nout"))
		no_out = 1;

	current.argv = argv + 1;
	for(;;)
	{
		c = long_getopt(sl_long_options, &current);
		switch(c)
		{
			case -1:
				if(!*current.argv)
				{
					fputs("No script file name given\n", stderr);
					usage(*argv);
					exit(EXIT_FAILURE);
				}
				program_argv = current.argv;
				program_argc = argc - (current.argv - argv);
				return;
			case 'c':
 				if(outs || ins)
					fatal_error("can't specify input/output streams with '-c' switch");
				as_cgi = 1;
				break;
			case '?':
				fprintf(stderr, "Invalid option: '%s'\n", unknown_option(&current));
			case 'h':
			case 'u':
				usage(*argv);
				exit(EXIT_FAILURE);
			case 'V':
				print_version();
				exit(0);
			case 'i':
				if(as_cgi)
					fatal_error("can't specify input/output streams with '-c' switch");
				ins = basic_stream_class.open(NULL, "stdin", 
											  *current.arg == '-' && !current.arg[1] ? "" : current.arg, 0);
				break;
			case 'o':
				if(!strcmp(optarg, "none"))
					no_out = 1;
				else
				{
					if(as_cgi)
						fatal_error("can't specify input/output streams with '-c' switch");	
					else outs = basic_stream_class.open(NULL, "stdout", *current.arg == '-' && !current.arg[1] ? "" : current.arg, 1);
				}
				break;
			case 'g':
				ds_reserved_mem_limit = get_amount(current.arg);
				break;
			case 'G':
				ds_unfix_mem_limit = get_amount(current.arg);
				break;
			case 'x':
				ds_unfix_max_cnt = strtoul(current.arg, NULL, 10);
				if(ds_unfix_max_cnt <= 0)
					fatal_error("invalid garbage counter");
				break;
			case 'd':
				def_mod = xstrdup(current.arg);
				break;
			case 'l':
				location = xstrdup(current.arg);
				break;
			case 's':
				add_script_path = xstrdup(current.arg);
				break;
			case 'f':
				flat_layout = 1;
				break;
			case 'S':
				syntax_only = 1;
				break;
		}
	}
}

static void setup_locations(void)
{
	dstring *temp = ds_create(location);
	dstring *temp1 = ds_create(NULL);
	if(DS_BODY(temp)[ds_length(temp) - 1] != '/')
		ds_appendch(temp, '/');
	nsl_location = xstrdup(DS_BODY(temp));
	if(flat_layout)
	{
		setenv(SL_TABLE_RO_PATH_ENV, DS_BODY(temp), 0);
		if(as_cgi)
			ds_appendstr(temp1, "./:");
		if(add_script_path)
		{
			ds_appendstr(temp1, add_script_path);
			ds_appendch(temp1, ':');
		}
		ds_append(temp1, temp);
		setenv(SCRIPT_PATH_ENV, DS_BODY(temp1), 0);
		setenv(AU_PATH_ENV, DS_BODY(temp1), 0);
	}
	else
	{
		if(as_cgi)
			ds_appendstr(temp1, "./:");
		if(add_script_path)
		{
			ds_appendstr(temp1, add_script_path);
			ds_appendch(temp1, ':');
		}
		ds_append(temp1, temp);
		ds_appendstr(temp1, "scripts:");
		ds_append(temp1, temp);
		ds_appendstr(temp1, "test");
		ds_appendch(temp1, ':');
		setenv(SCRIPT_PATH_ENV, DS_BODY(temp1), 0);
		ds_assign(temp1, temp);
		ds_appendstr(temp1, "tables");
		setenv(SL_TABLE_RO_PATH_ENV, DS_BODY(temp1), 0);
		ds_assign(temp1, temp);
		ds_appendstr(temp1, "automata:");
		ds_append(temp1, temp);
		ds_appendstr(temp1, "test");
		if(add_script_path)
		{
			ds_appendch(temp1, ':');
			ds_appendstr(temp1, add_script_path);
		}
		setenv(AU_PATH_ENV, DS_BODY(temp1), 0);
	}
}

int main(int argc, char *argv[])
{
	int exit_code;
	interpreter *i;
	shared_int *toplevel;

	parse_options(argc, argv);
	setup_locations();
	if(!as_cgi)
	{
		if(!outs && !no_out)
			outs = basic_stream_class.open(NULL, "stdout", "", 1);
	}
	else
	{
		ins = incgi_stream_class.open(NULL, "stdin", "", 0);
		if(!no_out)
			outs = outcgi_stream_class.open(NULL, "stdout", "", 1);
	}
#ifdef USE_LTDL
	init_modules(def_mod);
#endif
	i = load_interpreter(*program_argv, ins, outs);
	toplevel = init_shared_object(i, i, "subint", "");
	export_shared(toplevel->so, "toplevel");
	register_error_handler(NULL);
	if(CATCH())
	{
		if(!as_cgi)
			print_error_msg("run-time error: %s", get_error_desc());
		else
		{
			puts("Content-Type: text/plain\n");
			fprintf(stdout, "SL/CGI (%s): Run-time error: %s at %s:%d\n", *program_argv, get_error_desc(), 
					set_current_filename(NULL), set_line_no(-1));
		}
		return 1;
	}
	if(!syntax_only)
	{
  	 	exit_code = script_main_loop(i);	
		cleanup_interpreter(i);
	}
	return exit_code;
}
